#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Metric model
"""

#  from meerschaum.utils.misc import attempt_import
#  pydantic = attempt_import('pydantic')

#  class Metric(pydantic.BaseModel):
    #  connector_keys : str
    #  metric_key : str
    #  metric_name : str
