# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Import the pages.
"""

import meerschaum.api.dash.pages.error
import meerschaum.api.dash.pages.login
import meerschaum.api.dash.pages.dashboard
import meerschaum.api.dash.pages.plugins
import meerschaum.api.dash.pages.register
