#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define callbacks for pages.
"""

import meerschaum.api.dash.callbacks.dashboard
import meerschaum.api.dash.callbacks.login
import meerschaum.api.dash.callbacks.plugins
import meerschaum.api.dash.callbacks.jobs
import meerschaum.api.dash.callbacks.register
